@@ -2,7 +2,7 @@ package HTML::Selector::XPath;
 
 use strict;
 use 5.008_001;
-our $VERSION = '0.07';
+our $VERSION = '0.08';
 
 require Exporter;
 our @EXPORT_OK = qw(selector_to_xpath);
@@ -21,12 +21,12 @@ my $reg = {
     attr1   => qr/^\[([^\]]*)\]/,
     # attribute value match
     attr2   => qr/^\[\s*([^*~\|=\s:^\$]+)\s*([~\|*^\$]?=)\s*"([^"]+)"\s*\]/i,
-    attrN   => qr/^:not\((.*?)\)/i,
-    pseudo  => qr/^:([()a-z0-9_-]+)/i,
+    attrN   => qr/^:not\((.*?)\)/i, # this should be a parentheses matcher instead of a RE!
+    pseudo  => qr/^:([()a-z0-9_+-]+)/i,
     # adjacency/direct descendance
-    combinator => qr/^(\s*[>+~\s])/i,
+    combinator => qr/^(\s*[>+~\s](?!,))/i,
     # rule separator
-    comma => qr/^\s*,/i,
+    comma => qr/^\s*,\s*/i,
 };
 
 
@@ -61,6 +61,35 @@ sub convert_attribute_match {
     }
 };
 
+sub _generate_child {
+    my ($direction,$a,$b) = @_;
+    if ($a == 0) { # 0n+b
+        $b--;
+        "[count($direction-sibling::*) = $b and parent::*]"
+    } elsif ($a > 0) { # an + b
+        return "[not((count($direction-sibling::*)+1)<$b) and ((count($direction-sibling::*) + 1) - $b) mod $a = 0 and parent::*]"
+    } else { # -an + $b
+        $a = -$a;
+        return "[not((count($direction-sibling::*)+1)>$b) and (($b - (count($direction-sibling::*) + 1)) mod $a) = 0 and parent::*]"
+    };
+};
+
+sub nth_child {
+    my ($a,$b) = @_;
+    if (@_ == 1) {
+        ($a,$b) = (0,$a);
+    };
+    _generate_child('preceding', $a, $b);
+};
+
+sub nth_last_child {
+    my ($a,$b) = @_;
+    if (@_ == 1) {
+        ($a,$b) = (0,$a);
+    };
+    _generate_child('following', $a, $b);
+};
+
 sub to_xpath {
     my $self = shift;
     my $rule = $self->{expression} or return;
@@ -74,6 +103,7 @@ sub to_xpath {
     my $tag;
     my $wrote_tag;
     my $tag_index;
+    my $root_index = 0; # points to the current root
     # Loop through each "unit" of the rule
     while (length $rule && $rule ne $last_rule) {
         $last_rule = $rule;
@@ -135,23 +165,33 @@ sub to_xpath {
             } elsif ($sub_rule =~ s/$reg->{attr1}//) {
                 push @parts, "[not(\@$1)]";
             } else {
-                Carp::croak "Can't translate '$sub_rule' inside :not()";
+                my $xpath = selector_to_xpath($sub_rule);
+                $xpath =~ s!^//!!;
+                push @parts, "[not(self::$xpath)]";
+            #} else {
+            #    Carp::croak "Can't translate '$sub_rule' inside :not()";
             }
         }
 
         # Ignore pseudoclasses/pseudoelements
         while ($rule =~ s/$reg->{pseudo}//) {
             if ( $1 eq 'first-child') {
-                #$parts[$#parts] = '*[1]/self::' . $parts[$#parts];
-                # Replace the start of our current rule with a rule
-                # enforcing the current child
-                $parts[$tag_index] = '*[1]/self::' . $parts[$tag_index];
+                # Translates to :nth-child(1)
+                push @parts, nth_child(1);
             } elsif ( $1 eq 'last-child') {
-                push @parts, '[not(following-sibling::*)]';
+                push @parts, nth_last_child(1);
+            } elsif ( $1 eq 'only-child') {
+                push @parts, nth_child(1), nth_last_child(1);
             } elsif ($1 =~ /^lang\(([\w\-]+)\)$/) {
                 push @parts, "[\@xml:lang='$1' or starts-with(\@xml:lang, '$1-')]";
             } elsif ($1 =~ /^nth-child\((\d+)\)$/) {
-                push @parts, "[count(preceding-sibling::*) = @{[ $1 - 1 ]}]";
+                push @parts, nth_child($1);
+            } elsif ($1 =~ /^nth-child\((\d+)n(?:\+(\d+))?\)$/) {
+                push @parts, nth_child($1, $2||0);
+            } elsif ($1 =~ /^nth-last-child\((\d+)\)$/) {
+                push @parts, nth_last_child($1);
+            } elsif ($1 =~ /^nth-last-child\((\d+)n(?:\+(\d+))?\)$/) {
+                push @parts, nth_last_child($1, $2||0);
             } elsif ($1 =~ /^first-of-type$/) {
                 push @parts, "[1]";
             } elsif ($1 =~ /^nth-of-type\((\d+)\)$/) {
@@ -162,7 +202,7 @@ sub to_xpath {
                 push @parts, qq{[text()[contains(string(.),"$1")]]};
             } elsif ( $1 eq 'root') {
                 # This will give surprising results if you do E > F:root
-                $parts[0] = "/";
+                $parts[$root_index] = $root;
             } elsif ( $1 eq 'empty') {
                 push @parts, "[not(* or text())]";
             } else {
@@ -170,7 +210,7 @@ sub to_xpath {
             }
         }
 
-        # Match combinators (>, + and ~)
+        # Match combinators (whitespace, >, + and ~)
         if ($rule =~ s/$reg->{combinator}//) {
             my $match = $1;
             if ($match =~ />/) {
@@ -180,8 +220,10 @@ sub to_xpath {
                 $tag_index = $#parts;
             } elsif ($match =~ /\~/) {
                 push @parts, "/following-sibling::";
+            } elsif ($match =~ /^\s*$/) {
+                push @parts, "//"
             } else {
-                push @parts, "//";
+                die "Weird combinator '$match'"
             }
 
             # new context
@@ -192,6 +234,7 @@ sub to_xpath {
         # Match commas
         if ($rule =~ s/$reg->{comma}//) {
             push @parts, " | ", "$root/"; # ending one rule and beginning another
+            $root_index = $#parts;
             undef $tag;
             undef $wrote_tag;
         }
@@ -220,6 +263,9 @@ HTML::Selector::XPath - CSS Selector to XPath compiler
   my $relative = selector_to_xpath('div.foo', root => '/html/body/p' );
   # /html/body/p/div[contains(concat(' ', @class, ' '), ' foo ')]
 
+  my $relative = selector_to_xpath('div:root', root => '/html/body/p' );
+  # /html/body/p/div
+
 =head1 DESCRIPTION
 
 HTML::Selector::XPath is a utility function to compile full set of
@@ -281,6 +327,14 @@ patch to Prototype.js on L<http://dev.rubyonrails.org/ticket/5171>,
 but slightly modified using Aristotle Pegaltzis' CSS to XPath
 translation table per L<http://plasmasturm.org/log/444/>
 
+Also see
+
+L<http://www.mail-archive.com/www-archive@w3.org/msg00906.html>
+
+and
+
+L<http://kilianvalkhof.com/2008/css-xhtml/the-css3-not-selector/>
+
 =head1 LICENSE
 
 This library is free software; you can redistribute it and/or modify
